function [bestdesign, f0, fstar, R0, Rstar, fnKn, normKn, amspe, ...
    amspe0, amspeU, amspeU0, amspein, amspein0, omega, lossingenerations, n, N0, N, eps, model, params] = ...
    genetic(n, N0, N, eps, model, params, nx, unchangedLimit, ...
    popSize, crossoverProb, initMutationProb, finalMutationProb, eliteProp, indesign)



%% INPUT:
% n = sample size
% N = # of time points
% N0 = sampling upper limit
% model = correlation model
% params = [phi1, phi1 phi2, theta1, theta1 theta2] for AR1, AR2, MA1, MA2
    % models
% nx = # of grid points for Simpson's Rule; must be odd
% intercept = 1 to fit an intercept; = 0 otherwise
% makeplots = 1 to plot output; = 0 otherwise
% unchangedLimit = number of consecutive generations with the same best
    % design, for termination
% popSize = population size
% crossoverProb = prob that two chosen parents will reproduce; otherwise
    % the child is a copy of the best of the parents
% init/final mutationProb = inital and final prob that the child will undergo a random mutation, in
    % each of n randomly chosen spots; varies linearly from init to final with "unchanged"
% eliteProp = proportion of population (the best members) to always survive
% indesign = [], or enter a particular design here

%% OUTPUT:
% f0 = base spectral density
% fstar = least favourable
% R0 = base acf
% Rstar = least favourable
% fnKn = Kn(omega)
% normKn = its norm
% ampse = minimax amspe
% amspe0 = amspe of this deisgn at the base model (eps = 0)
% amspeU = max amspe of the unikfrom design
% ampseU0 = amspe of the uniform at the base model
% the rest is obvious
%% examples: 
%% [bestdesign, f0, fstar, R0, Rstar, fnKn, normKn, amspe, amspe0, amspeU, amspeU0, amspein, amspein0, omega, lossingenerations, n, N0, N, eps, model, params] =
%Fig 1:  genetic(10, 100, 100, .5, 'AR1', [-.5 1 -.5  -.9  -1 .6], 1001, 1000, 40, .95, .2, 0, .2, []); % save ar1.mat
%Fig 2:  genetic(10, 75,  100, .25, 'MA1', [-.5 1 -.5  -.9  -1 .6], 1001,1000, 40, .95, .2, 0, .2, []); % save ma1.mat 


format short g
tStart = tic;

GENMAT = []; % Keep track of the best set of design points in each generation
POPULATION = zeros(popSize, n); % Population of designs represented as time points
generation = 1; 
unchanged = 0;

% grid for Simpsons's rule: 
omega = linspace(-pi, pi, nx);
Mx = nx-1;
hx = 2*pi/Mx;
dx = [1;repmat([4;2], [Mx/2-1,1]);4;1];
kx = hx*dx/3;

%%================================%%
% CREATE f0 and A here

if strcmp(model,'AR1')
    params = params(1);
    phi = params;
    f0 = (1 - phi^2)./(2*pi*(1 + phi^2 - 2*phi*cos(omega))); % a row
    seed = [5,12,19,26,33,40,48,56,65,75];
end

if strcmp(model,'AR2')
    params = params(2:3);
    phi1 = params(1);
    phi2 = params(2);
    num = (1 + phi2)*((1 - phi2)^2 - phi1^2)/(2*pi*(1 - phi2));
    f0 = num./(1 + phi1^2 + phi2^2 - 2*phi1*(1 - phi2)*cos(omega) - 2*phi2*(cos(2*omega))); % a row
    seed = [8,15,25,36,43,54,64,74,85,92];
end

if strcmp(model,'MA1')
    params = params(4);
    theta = params;
    f0 = (1 + theta^2 + 2*theta*cos(omega))/(2*pi*(1 + theta^2)); % a row
    seed = [4,7,20,25,40,47,59,67,72,74];
end

if strcmp(model,'MA2')
    params = params(5:6);
    theta1 = params(1);
    theta2 = params(2);
    f0 = (1 + theta1^2 + theta2^2 + 2*theta1*(1 + theta2)*cos(omega) + 2*theta2*cos(2*omega))/ ...
        (2*pi*(1 + theta1^2 + theta2^2)); % a row 
    seed = [15,36,41,51,56,68,75,81,85,93];
end

if strcmp(model,'ARMA11')
    params = [params(1), params(4)];
    phi = params(1); %ar param
    theta = params(2); %ma param
    f0 = (1 + theta^2 + 2*theta*cos(omega))./(1 + phi^2 - 2*phi*cos(omega)); % a row 
    f0 = f0/(f0*kx);
    seed = [];
end

variance = f0*kx;
normf0 = sqrt((f0.^2)*kx);
ARR = zeros(N,nx);
for j=1:N
	ARR(j,:) = cos((j-1)*omega).*f0;
end
correlations = ARR*kx;
CORR = toeplitz(correlations); 

%% Simulate an initial population of designs; compute the loss for each member

for count = 1:popSize % or 1000, then uncomment the lines below and 
        %choose the BEST popSize of these randomly chosen designs
	des = randperm(N0);
	des = sort(des(1:n));
    POPULATION(count,:) = des;
end
if length(seed) == n
    POPULATION(1,:) = seed;
end

lossvec = loss(POPULATION, CORR, eps);
fitvec = fitness(lossvec); % Fitness of all designs in this generation


%% Create subsequent generations
while unchanged < unchangedLimit
    newPopulation = zeros(popSize, n);
    % Force elites (the best of the current population) to survive
    numberElites = floor(popSize*eliteProp);
    [~,I] = sort(fitvec,'descend');
    newPopulation(1:numberElites, :) = POPULATION(I(1:numberElites),:);
    lossvec(1:numberElites) = lossvec(I(1:numberElites));
    lossvec(numberElites:popSize) = zeros(popSize - numberElites + 1,1);
    newPopSize = numberElites;
    
    % Create the remaining children:
    cumfit = cumsum(fitvec);    
    while(newPopSize < popSize)
        % Create a child from 2 randomly chosen members of the current population        
        mutationProb = initMutationProb + ...
            (finalMutationProb-initMutationProb)*unchanged/unchangedLimit;
        child  = combine(POPULATION, fitvec, cumfit, n, N0, crossoverProb, mutationProb);
        % Update the new population
        newPopSize = newPopSize + 1;
        newPopulation(newPopSize,:) = child;
    end % end of inner "while"     
    POPULATION = newPopulation;  % Replace the old population with the new one
    CHILDREN = POPULATION(numberElites:popSize,:);
    lossvec(numberElites:popSize) = loss(CHILDREN, CORR, eps);

    % Summarize this generation:
    bestdesign = POPULATION(lossvec == min(lossvec), :);
    if size(bestdesign,1) > 1
        bestdesign = bestdesign(1, :);
    end
    GENMAT = [GENMAT; [bestdesign, min(lossvec)]];
    if generation > 1 && isequal(GENMAT(generation, 1:n), GENMAT(generation - 1, 1:n)) % restart when the best design changes
%	if generation > 1 && isequal(GENMAT(generation, n+1), GENMAT(generation - 1, n+1)) % restart when the min loss changes
        unchanged = unchanged + 1;
    else
        unchanged = 0;
    end
	fitvec = fitness(lossvec); % fitness of all designs in generation
    disp('     generation, unchanged, mutationProb and minloss:');
    disp([generation, unchanged, mutationProb, min(lossvec)]);
	generation = generation + 1;
end % end of outer "while"


%% Summarize and plot the output
bestDesignInGenerations = GENMAT(:, 1:n);
bestLossInGenerations = GENMAT(:, n+1);
bestdesign = POPULATION(lossvec == min(lossvec), :);
if size(bestdesign,1) > 1
	bestdesign = bestdesign(1, :);
end

desIndices = zeros(1,N0);
desIndices(bestdesign) = 1;

fnKn  = Kn(omega, bestdesign, n, N, CORR);
normKn  = normK(bestdesign, n, N, CORR);
fstar = f0 + eps*fnKn/normKn;
R0 = CORR(:,1);
Rstar  = R(bestdesign, n, N, CORR, eps);
amspe = min(lossvec);
varfstar = fstar*kx;
lossingenerations = GENMAT(:,n+1);
figure(1);

subplot(2,2,1);
plot(lossingenerations);
string0 = ['AMSPE = ', num2str(myround(amspe, 3))]; 
legend(string0);
xlabel('(a) Minloss vs. generation');

subplot(2,2,2);
bar(desIndices, .01);
xlim([0 N0+1]);
xlabel('(b) Best design')

subplot(2,2,3);
plot(omega, fstar, '-r', omega, f0, '-.b');
xlim([-pi pi]);
legend('f_{**}','f_{0}');
xlabel('(c) spectral densities');

subplot(2,2,4);
last = 7;
plot(0:(last-1), Rstar(1:last)', '-r', 0:(last-1), R0(1:last), '-.b');
xlim([0 last - .5]);
ylim([-1 1]);
legend('R_{**}','R_{0}');
xlabel('(d) autocorrelation functions')

if strcmp(model,'AR1') || strcmp(model,'MA1')
    string1 = [ 'n  = ', num2str(n), ' N_{0} = ', num2str(N0), ' N = ', num2str(N), ' eps = ', num2str(eps)];
    string2 = ['Base model is ', char(model), ' with parameter ', ...
        num2str(params), '; ||f_{0}|| = ', num2str(normf0)];
    string3 = ['Best design is ', num2str(bestdesign)];
end

if strcmp(model,'AR2') || strcmp(model,'MA2')
    string1 = [ 'n  = ', num2str(n), ' N_{0} = ', num2str(N0), ' N = ', num2str(N), ' eps = ', num2str(eps)];
    string2 = ['Base model is ', char(model), ' with parameters ', ...
        num2str(params(1)), ' and ', num2str(params(2)), '; ||f_{0}|| = ', num2str(normf0)];
    string3 = ['Best design is ', num2str(bestdesign)];
end



tElapsed = toc(tStart);
tElapsed = myround(tElapsed/60, 3); % time in minutes

disp(['Model is ', model, ', with parameters ', num2str(params)]);
disp(['Elapsed time = ', num2str(tElapsed), ' minutes, # of generations = ', num2str(size(GENMAT, 1))]);
disp(['n = ', num2str(n), '  N0 = ', num2str(N0),'  N = ', num2str(N), '  eps = ', num2str(eps)]);
disp(['Varf0 = ', num2str(variance), '  Norm of f0 = ', num2str(normf0), '  Varf** = ', num2str(varfstar)]);
disp(['Best design is ', num2str(bestdesign)]);
disp(['Minmax AMSPE is ', num2str(amspe)]);


amspe0  = loss(bestdesign, CORR, 0);
disp(['AMSPE at base model is ', num2str(amspe0)]);

uniformdes = round(N0*((1:n)-.5*ones(1,n))/n);
disp(['uniform design is ', num2str(uniformdes)]);
amspeU = loss(uniformdes, CORR, eps); 
disp(['max AMSPE of uniform design is ', num2str(amspeU)]);
amspeU0 = loss(uniformdes, CORR, 0); 
disp(['AMSPE of uniform at base model is ', num2str(amspeU0)]);
if ~isempty(indesign)
    disp(['input design is ', num2str(indesign)]);
    amspein = loss(indesign, CORR, eps); 
    disp(['max AMSPE of input design is ', num2str(amspein)]);
    amspein0 = loss(indesign, CORR, 0); 
    disp(['AMSPE of input at base model is ', num2str(amspein0)]);
else
    amspein = [];
    amspein0 = [];
end

save thisoutput.mat 


%% ============================= subfunctions =============================
%% ========================================================================


%% ============================= subfunction "fitness" =============================
function fit = fitness(x) 
%fit = 1./(x - .99*min(x)).^2;
fit = 1./sqrt(tiedrank(x));
fit = fit/sum(fit);

%% ============================= subfunction "combine" =============================
function child  = combine(POPULATION, fitvec, cumfit, n, N, crossoverProb, mutationProb) 
popIndices = 1:size(POPULATION, 1);
parents = zeros(2, N);
desIndices = 1:N;
choices = zeros(2,1);

for i = 1:2
	choice = min(popIndices(cumfit >= rand));
    parent = POPULATION(choice,:);
    parents(i, parent) = 1;  % vector of N zeros and ones 
    choices(i) = choice;
end

if rand <= crossoverProb  
	% With this prob, form a child from the two designs chosen above
	child = parents(1,:) + parents(2,:);
	singletonsToDrop = length(child(child > 0)) - n; 
	% drop some points which appear only once:
        if singletonsToDrop > 0
            singletons = desIndices(child == 1);
            ss = randperm(length(singletons));
            singletons = singletons(ss); % permute the singletons
            indicesToDrop = singletons(1:singletonsToDrop);
            child(indicesToDrop) = 0;
        end
	child(child == 2) = 1; % keep any point which appears twice
else % otherwise the child is identical to the best of the parents
	if fitvec(choices(1)) > fitvec(choices(2))
        child = parents(1,:);
    else
        child = parents(2,:);
	end
end
numberOfMutations = 0;
while numberOfMutations <= n  % Possibly cause up to n mutations
	if rand < mutationProb
        s1 = randperm(n);
        s0 = randperm(N-n);
        child1 = desIndices(child==1);
        child0 = desIndices(child==0);
        child(child1(s1(1))) = 0;
        child(child0(s0(1))) = 1;    
	end
	numberOfMutations = numberOfMutations +1;
end
indi = 1:N;
child = indi(child == 1);
child = sort(child);


%% ============================= subfunction "loss" =============================
function out  = loss(POPULATION, CORR, eps)
n = size(POPULATION,2);
N = size(CORR,1);
popSize = size(POPULATION,1);
out = zeros(popSize,1);

arrayE  = E(POPULATION, n, N);
arrayA  = A(POPULATION, CORR, n, N);

% arrayE is an array of size (popSize by N by N-1 by n)
% E(m,k,t,j) = I(|k - (jth point of design m)| == t)
% For design m (m = 1,...,popSize):
% E_n,N(t) = permute(E(m,:,t,:), [4,2,1,3]) : n by N
% E_n(t) = [E_n,N(t)](:,POPULATION(m,:)) : n by n

% For design m (m = 1,...,popSize):
% arrayA = (a_n(1),...,a_n(N)): n by N
% A_n = permute(arrayA(m,:, POPULATION(m,:)), [2,3,1]) : n by n
% B_n,N = inverse(A_n)*arrayA : n by N

if eps>0
    for m=1:popSize
        Ebig = permute(permute(arrayE(m,:,:,:), [4,2,1,3]),[1,2,4,3]);
        %n by N by N-1; each n by N face is E_n,N(t) for some t = 1,...,N-1
        Esmall = Ebig(:,POPULATION(m,:),:);
        % n by n by N-1; each n by n face is E_n(t) for some t
        An = permute(arrayA(m,:, POPULATION(m,:)), [2,3,1]);
        BnN = An\permute(arrayA(m,:,:),[2,3,1]); % n by N
        BBprime = BnN*BnN';
        s = zeros(N-1,1);
        for t = 1:(N-1)
            s(t) = trace(Esmall(:,:,t)*BBprime) - 2*sum(sum(Ebig(:,:,t)*BnN'));
        end
        %   normKn(m) = sqrt(pi)*norm(s)/N
        out(m) = 1 - trace(BBprime*An)/N + eps*sqrt(pi)*norm(s)/N;
    end
else
    for m=1:popSize
        An = permute(arrayA(m,:, POPULATION(m,:)), [2,3,1]);
        BnN = An\permute(arrayA(m,:,:),[2,3,1]); % n by N
        BBprime = BnN*BnN';
        out(m) = 1 - trace(BBprime*An)/N;
    end
end

%% ============================= subfunction "fnKn" =============================
function fnKn  = Kn(omega, des, n, N, CORR) % row of values of Kn(omega)

BnN = CORR(des,des)\CORR(des,:);

nx = length(omega);
fnKn = zeros(1,nx);
Mn = zeros(n, n, nx);
PnN = zeros(n, N, nx);

for l = 1:nx
    for j = 1:n
        for k = 1:n
            Mn(j,k,l) = cos(omega(l)*(abs(des(j)-des(k))));            
        end
    end
end

for l = 1:nx
    for j = 1:n
        for k = 1:N
            PnN(j,k,l) = cos(omega(l)*(abs(des(j)-k)));
        end
    end
end  

for l = 1:nx
    fnKn(l) = ((trace(Mn(:,:,l)*(BnN*BnN'))-sum(sum(BnN.^2))) - 2*(trace(PnN(:,:,l)*BnN')-n))/N;
end

%% ============================= subfunction "normK" =============================
function normKn  = normK(des, n, N, CORR) 
BnN = CORR(des,des)\CORR(des,:);
sumsq = 0;
for t=1:(N-1)
	EnN = E1(t, des, n, N);
	En = EnN(:,des);
	losst = (trace(En*(BnN*BnN')) - 2*sum(sum(EnN*BnN')))^2;
	sumsq = sumsq + losst;
end
normKn = sqrt(pi*sumsq)/N;

%% ============================= subfunction "R" =============================
function Rstar  = R(des, n, N, CORR, eps) 

Rstar = zeros(1,N);
normKn  = normK(des, n, N, CORR);
BnN = CORR(des,des)\CORR(des,:);

for s = 2:N
	EnN = E1(s-1, des, n, N);
	En = EnN(:,des);
	Rstar(s) = trace(En*(BnN*BnN')) - 2*trace(EnN*BnN');
end

Rstar = CORR(1,:) + eps*pi*Rstar/(N*normKn);


%% ============================= subfunction "E" =============================
function arrayE  = E(POPULATION, n, N)
arrayE = zeros(size(POPULATION,1), N, N-1, n);
for t = 1:(N-1)
    for k = 1:N
        arrayE(:,k,t,:) = (abs(k-POPULATION) == t);
    end
end
   

%% ============================= subfunction "A" =============================
function arrayA  = A(POPULATION, CORR, n, N)
popSize = size(POPULATION,1);
arrayA = zeros(popSize, n, N);
for m = 1:popSize
    arrayA(m,:,:) = CORR(POPULATION(m,:),1:N);
end


%% ============================= subfunction "E1" =============================
function mat  = E1(s, des, n, N) 

mat = zeros(n,N);
for j = 1:n
    for k = 1:N
        mat(j,k) = (abs(des(j)-k)==s);
    end
end    

%% ============================= subfunction "myround" =============================
function yy = myround(x, d)
yy = round(x*10^d)/10^d;

