function [bestdesign, f0, fstar, R0, Rstar, fnKn, normKn, amspe, amspe0, amspeU, amspeU0, amspein, amspein0, omega, lossingenerations, n, N0, N, eps, model, params] = ...
    longgenetic(n, N0, N, eps, model, params, nx, unchangedLimit, popSize, crossoverProb, initMutationProb, finalMutationProb, eliteProp, indesign)



%% INPUT:
% n = sample size
% N = # of time points
% N0 = sampling upper limit
% model = correlation model
% params = [phi1, phi1 phi2, theta1, theta1 theta2] for AR1, AR2, MA1, MA2
    % models
% nx = # of grid points for Simpson's Rule; must be odd
% intercept = 1 to fit an intercept; = 0 otherwise
% makeplots = 1 to plot output; = 0 otherwise
% unchangedLimit = number of consecutive generations with the same best
    % design, for termination
% popSize = population size
% crossoverProb = prob that two chosen parents will reproduce; otherwise
    % the child is a copy of the best of the parents
% init/final mutationProb = inital and final prob that the child will undergo a random mutation, in
    % each of n randomly chosen spots; varies linearly from init to final with "unchanged"
% eliteProp = proportion of population (the best members) to always survive
% indesign = [], or enter a particular design here

%% OUTPUT:
% f0 = base spectral density
% fstar = least favourable
% R0 = base acf
% Rstar = least favourable
% fnKn = Kn(omega)
% normKn = its norm
% ampse = minimax amspe
% amspe0 = amspe of this deisgn at the base model (eps = 0)
% amspeU = max amspe of the unikfrom design
% ampseU0 = amspe of the uniform at the base model
% the rest is obvious
%% examples: 
%% [bestdesign, f0, fstar, R0, Rstar, fnKn, normKn, amspe, amspe0, amspeU, amspeU0, amspein, amspein0, omega, lossingenerations, n, N0, N, eps, model, params] =
%Fig 1:  longgenetic(10, 100, 100, .5, 'AR1', [-.5 1 -.5  -.9  -1 .6], 1001, 1000, 40, .95, .2, 0, .2, []); % save ar1.mat
%Fig 2:  longgenetic(10, 75,  100, .25, 'MA1', [-.5 1 -.5  -.9  -1 .6], 1001,1000, 40, .95, .2, 0, .2, []); % save ma1.mat 


format short g
tStart = tic;

lossvec = zeros(popSize,1); %loss of each member of the population in one generation
GENMAT = []; % Keep track of the best set of design points in each generation
POPULATION = zeros(popSize, n); % Population of designs represented as time points
generation = 1; 
unchanged = 0;
NEG = []; % Will record designs for which the fstar of Thoerem 2 becomes negative

% grid for Simpsons's rule: 
omega = linspace(-pi, pi, nx);
Mx = nx-1;
hx = 2*pi/Mx;
dx = [1;repmat([4;2], [Mx/2-1,1]);4;1];
kx = hx*dx/3; % Integral of f over [-pi,pi] is f*kx

%%================================%%
% CREATE f0 and A here

if strcmp(model,'AR1')
    params = params(1);
    phi = params;
    f0 = (1 - phi^2)./(2*pi*(1 + phi^2 - 2*phi*cos(omega))); % a row   
end

if strcmp(model,'AR2')
    params = params(2:3);
    phi1 = params(1);
    phi2 = params(2);
    num = (1 + phi2)*((1 - phi2)^2 - phi1^2)/(2*pi*(1 - phi2));
    f0 = num./(1 + phi1^2 + phi2^2 - 2*phi1*(1 - phi2)*cos(omega) - 2*phi2*(cos(2*omega))); % a row
end

if strcmp(model,'MA1')
    params = params(4);
    theta = params;
    f0 = (1 + theta^2 + 2*theta*cos(omega))/(2*pi*(1 + theta^2)); % a row   
end

if strcmp(model,'MA2')
    params = params(5:6);
    theta1 = params(1);
    theta2 = params(2);
    f0 = (1 + theta1^2 + theta2^2 + 2*theta1*(1 + theta2)*cos(omega) + 2*theta2*cos(2*omega))/ ...
        (2*pi*(1 + theta1^2 + theta2^2)); % a row   
end

if strcmp(model,'ARMA11')
    params = [params(1), params(4)];
    phi = params(1); %ar param
    theta = params(2); %ma param
    f0 = (1 + theta^2 + 2*theta*cos(omega))./(1 + phi^2 - 2*phi*cos(omega)); % a row 
    f0 = f0/(f0*kx);
end

variance = f0*kx;
normf0 = sqrt((f0.^2)*kx);
ARR = zeros(N,nx);
for j=1:N
	ARR(j,:) = cos((j-1)*omega).*f0;
end
correlations = ARR*kx;
A = toeplitz(correlations); 

%% Simulate an initial population of designs; compute the loss for each member

for count = 1:popSize % or 1000, then uncomment the lines below and 
        %choose the BEST popSize of these randomly chosen designs
	des = randperm(N0);
	des = sort(des(1:n));
    [lossvec(count), NEG] = loss(des, n, N, A, eps, omega, f0, NEG);
    POPULATION(count,:) = des;
end

fitvec = fitness(lossvec); % Fitness of all designs in this generation


%% Create subsequent generations
while unchanged < unchangedLimit
    newPopulation = zeros(popSize, n);
    % Force elites (the best of the current population) to survive
    numberElites = floor(popSize*eliteProp);
    [~,I] = sort(fitvec,'descend');
    newPopulation(1:numberElites, :) = POPULATION(I(1:numberElites),:);
    lossvec(1:numberElites) = lossvec(I(1:numberElites));
    newPopSize = numberElites;
    
    % Create the remaining children:
    cumfit = cumsum(fitvec);    
    while(newPopSize < popSize)
        % Create a child from 2 randomly chosen members of the current population        
        mutationProb = initMutationProb + ...
            (finalMutationProb-initMutationProb)*unchanged/unchangedLimit;
        child  = combine(POPULATION, fitvec, cumfit, n, N0, N, crossoverProb, mutationProb);
        % Update the new population
        newPopSize = newPopSize + 1;
        newPopulation(newPopSize,:) = child;
        % Compute the loss
        [lossvec(newPopSize), NEG]  = loss(child, n, N, A, eps, omega, f0, NEG);
    end % end of inner "while"     
    POPULATION = newPopulation;  % Replace the old population with the new one
    % Summarize this generation:
    bestdesign = POPULATION(lossvec == min(lossvec), :);
    if size(bestdesign,1) > 1
        bestdesign = bestdesign(1, :);
    end
    GENMAT = [GENMAT; [bestdesign, min(lossvec)]];
    if generation > 1 && isequal(GENMAT(generation, 1:n), GENMAT(generation - 1, 1:n)) % restart when the best design changes
%	if generation > 1 && isequal(GENMAT(generation, n+1), GENMAT(generation - 1, n+1)) % restart when the min loss changes
        unchanged = unchanged + 1;
    else
        unchanged = 0;
    end
	fitvec = fitness(lossvec); % fitness of all designs in generation
    disp('     generation, unchanged, mutationProb and minloss:');
    disp([generation, unchanged, mutationProb, min(lossvec)]);
	generation = generation + 1;
end % end of outer "while"


%% Summarize and plot the output
bestDesignInGenerations = GENMAT(:, 1:n);
bestLossInGenerations = GENMAT(:, n+1);
bestdesign = POPULATION(lossvec == min(lossvec), :);
if size(bestdesign,1) > 1
	bestdesign = bestdesign(1, :);
end

desIndices = zeros(1,N0);
desIndices(bestdesign) = 1;

fnKn  = Kn(omega, bestdesign, n, N, A);
normKn  = normK(bestdesign, n, N, A);
fstar = f0 + eps*fnKn/normKn;
R0 = A(:,1);
Rstar  = R(bestdesign, n, N, A, eps);
amspe = min(lossvec);
varfstar = fstar*kx;
lossingenerations = GENMAT(:,n+1);
figure(1);

subplot(2,2,1);
plot(lossingenerations);
string0 = ['AMSPE = ', num2str(myround(amspe, 3))]; 
legend(string0);
xlabel('(a) Minloss vs. generation');

subplot(2,2,2);
bar(desIndices, .01);
xlim([0 N0+1]);
xlabel('(b) Best design')

subplot(2,2,3);
plot(omega, fstar, '-r', omega, f0, '-.b');
xlim([-pi pi]);
legend('f_{**}','f_{0}');
xlabel('(c) spectral densities');

subplot(2,2,4);
last = 7;
plot(0:(last-1), Rstar(1:last)', '-r', 0:(last-1), R0(1:last), '-.b');
xlim([0 last - .5]);
ylim([-1 1]);
legend('R_{**}','R_{0}');
xlabel('(d) autocorrelation functions')

if strcmp(model,'AR1') || strcmp(model,'MA1')
    string1 = [ 'n  = ', num2str(n), ' N_{0} = ', num2str(N0), ' N = ', num2str(N), ' eps = ', num2str(eps)];
    string2 = ['Base model is ', char(model), ' with parameter ', ...
        num2str(params), '; ||f_{0}|| = ', num2str(normf0)];
    string3 = ['Best design is ', num2str(bestdesign)];
end

if strcmp(model,'AR2') || strcmp(model,'MA2')
    string1 = [ 'n  = ', num2str(n), ' N_{0} = ', num2str(N0), ' N = ', num2str(N), ' eps = ', num2str(eps)];
    string2 = ['Base model is ', char(model), ' with parameters ', ...
        num2str(params(1)), ' and ', num2str(params(2)), '; ||f_{0}|| = ', num2str(normf0)];
    string3 = ['Best design is ', num2str(bestdesign)];
end

string0 = ['    '];

%suptitle({string0; string1; string2;string3;string0});

legend('R_{**}','R_{0}');

tElapsed = toc(tStart);
tElapsed = myround(tElapsed/60, 3); % time in minutes

disp(['Model is ', model, ', with parameters ', num2str(params)]);
disp(['Elapsed time = ', num2str(tElapsed), ' minutes, # of generations = ', num2str(size(GENMAT, 1))]);
disp(['n = ', num2str(n), '  N0 = ', num2str(N0),'  N = ', num2str(N), '  eps = ', num2str(eps)]);
disp(['Varf0 = ', num2str(variance), '  Norm of f0 = ', num2str(normf0), '  Varf** = ', num2str(varfstar)]);
disp(['Best design is ', num2str(bestdesign)]);
disp(['Minmax AMSPE is ', num2str(amspe)]);

[amspe0, NEG]  = loss(bestdesign, n, N, A, 0, omega, f0, NEG); 
disp(['AMSPE at base model is ', num2str(amspe0)]);

uniformdes = round(N0*((1:n)-.5*ones(1,n))/n);
disp(['uniform design is ', num2str(uniformdes)]);
[amspeU, NEG] = loss(uniformdes, n, N, A, eps, omega, f0, NEG); 
disp(['max AMSPE of uniform design is ', num2str(amspeU)]);
[amspeU0, NEG] = loss(uniformdes, n, N, A, 0, omega, f0, NEG); 
disp(['AMSPE of uniform at base model is ', num2str(amspeU0)]);
if ~isempty(indesign)
    disp(['input design is ', num2str(indesign)]);
    [amspein, NEG] = loss(indesign, n, N, A, eps, omega, f0, NEG); 
    disp(['max AMSPE of input design is ', num2str(amspein)]);
    [amspein0, NEG] = loss(indesign, n, N, A, 0, omega, f0, NEG); 
    disp(['AMSPE of input at base model is ', num2str(amspein0)]);
else
    amspein = [];
    amspein0 = [];
end

disp(['Number of designs for which fstar becomes negative is ', num2str(size(NEG,1))]);
disp(NEG)
save thisoutput.mat


%% ============================= subfunctions =============================
%% ========================================================================


%% ============================= subfunction "fitness" =============================
function fit = fitness(x) 
%fit = 1./(x - .99*min(x)).^2;
fit = 1./sqrt(tiedrank(x));
fit = fit/sum(fit);

%% ============================= subfunction "combine" =============================
function child  = combine(POPULATION, fitvec, cumfit, n, N0, N, crossoverProb, mutationProb) 
popIndices = 1:size(POPULATION, 1);
parents = zeros(2, N0);
desIndices = 1:N0;
choices = zeros(2,1);

for i = 1:2
	choice = min(popIndices(cumfit >= rand));
    parent = POPULATION(choice,:);
    parents(i, parent) = 1;  % vector of N0 zeros and ones 
    choices(i) = choice;
end

if rand <= crossoverProb  
	% With this prob, form a child from the two designs chosen above
	child = parents(1,:) + parents(2,:);
	singletonsToDrop = length(child(child > 0)) - n; 
	% drop some points which appear only once:
        if singletonsToDrop > 0
            singletons = desIndices(child == 1);
            ss = randperm(length(singletons));
            singletons = singletons(ss); % permute the singletons
            indicesToDrop = singletons(1:singletonsToDrop);
            child(indicesToDrop) = 0;
        end
	child(child == 2) = 1; % keep any point which appears twice
else % otherwise the child is identical to the best of the parents
	if fitvec(choices(1)) > fitvec(choices(2))
        child = parents(1,:);
    else
        child = parents(2,:);
	end
end
numberOfMutations = 0;
while numberOfMutations <= n  % Possibly cause up to n mutations
	if rand < mutationProb
        s1 = randperm(n);
        s0 = randperm(N0-n);
        child1 = desIndices(child==1);
        child0 = desIndices(child==0);
        child(child1(s1(1))) = 0;
        child(child0(s0(1))) = 1;    
	end
	numberOfMutations = numberOfMutations +1;
end
indi = 1:N;
child = indi(child == 1);
child = sort(child);


%% ============================= subfunction "loss" =============================
function [out, NEG]  = loss(des, n, N, A, eps, omega, f0, NEG) 

BnN = A(des,des)\A(des,:);
sumsq = 0;
normKn = 0;
fstar = f0;
if eps>0
    for t=1:(N-1)
        EnN = E(t, des, n, N);
        En = EnN(:,des);
        losst = (trace(En*(BnN*BnN')) - 2*sum(sum(EnN*BnN')))^2;
        sumsq = sumsq + losst;
    end
    fnKn  = Kn(omega, des, n, N, A);
    normKn = sqrt(pi*sumsq)/N;
    fstar = f0 + eps*fnKn/normKn;
end
if any(fstar < 0) 
    NEG = [NEG; des]
    disp(['Go to line 347 of the .m file and insert code to compute the ']); 
    disp(['loss using Theorem 1, rather than Theorem 2']);    
end
out = 1 - trace(BnN*A(des,:)')/N + eps*normKn;

%% ============================= subfunction "fnKn" =============================
function fnKn  = Kn(omega, des, n, N, A) % row of values of Kn(omega)

BnN = A(des,des)\A(des,:);

nx = length(omega);
fnKn = zeros(1,nx);
Mn = zeros(n, n, nx);
PnN = zeros(n, N, nx);

for l = 1:nx
    for j = 1:n
        for k = 1:n
            Mn(j,k,l) = cos(omega(l)*(abs(des(j)-des(k))));            
        end
    end
end

for l = 1:nx
    for j = 1:n
        for k = 1:N
            PnN(j,k,l) = cos(omega(l)*(abs(des(j)-k)));
        end
    end
end  

for l = 1:nx
    fnKn(l) = ((trace(Mn(:,:,l)*(BnN*BnN'))-sum(sum(BnN.^2))) - 2*(trace(PnN(:,:,l)*BnN')-n))/N;
end

%% ============================= subfunction "normK" =============================
function normKn  = normK(des, n, N, A) 

BnN = A(des,des)\A(des,:);
sumsq = 0;
for t=1:(N-1)
	EnN = E(t, des, n, N);
	En = EnN(:,des);
	losst = (trace(En*(BnN*BnN')) - 2*sum(sum(EnN*BnN')))^2;
	sumsq = sumsq + losst;
end
normKn = sqrt(pi*sumsq)/N;

%% ============================= subfunction "R" =============================
function Rstar  = R(des, n, N, A, eps) 

Rstar = zeros(1,N);
normKn  = normK(des, n, N, A);
BnN = A(des,des)\A(des,:);

for s = 2:N
	EnN = E(s-1, des, n, N);
	En = EnN(:,des);
	Rstar(s) = trace(En*(BnN*BnN')) - 2*trace(EnN*BnN');
end

Rstar = A(1,:) + eps*pi*Rstar/(N*normKn);


%% ============================= subfunction "E" =============================
function mat  = E(s, des, n, N) 

mat = zeros(n,N);
for j = 1:n
    for k = 1:N
        mat(j,k) = (abs(des(j)-k)==s);
    end
end    

%% ============================= subfunction "myround" =============================
function y = myround(x, d)
y = round(x*10^d)/10^d;

